import functools
from collections import namedtuple

METHOD_MAPPING = {
    "POST": ("create", "bulk_create"),
    "GET": ("read",),
    "PATCH": ("update", "bulk_update"),
    "DELETE": ("delete",),
}


def collect_routes(falcon_app):
    raw_routes = []
    for root_r in falcon_app._router._roots:
        for children_r in root_r.children:
            raw_routes.append(
                (
                    children_r.uri_template,
                    children_r.var_name,
                    children_r.method_map,
                    children_r.resource,
                )
            )

            for a in children_r.children:

                if a.children:
                    for a2 in a.children:
                        raw_routes.append(
                            (
                                a2.uri_template,
                                a2.var_name,
                                a2.method_map,
                                a2.resource,
                            )
                        )
                raw_routes.append(
                    (a.uri_template, a.var_name, a.method_map, a.resource)
                )

    routes = []
    new_route = namedtuple(
        "route", ["route", "uri", "method_name", "method", "resource"]
    )

    for r_name, var_name, methods, resource in raw_routes:
        if not methods:
            continue

        for method_name, method in methods.items():
            valid_methods = METHOD_MAPPING.get(method_name, [])
            if not any(m in resource.Meta.methods for m in valid_methods):
                continue

            if hasattr(resource.Meta, "skip_doc") and resource.Meta.skip_doc:
                continue

            if not isinstance(method, functools.partial):
                r = new_route(
                    r_name, var_name, method_name.lower(), method, resource
                )
                routes.append(r)

    return routes



import falcon

from .base import BaseApiException


class BadRequest(BaseApiException):
    def __init__(self, details="", code="bad-request"):
        BaseApiException.__init__(
            self,
            status=falcon.HTTP_BAD_REQUEST,
            title=falcon.HTTP_BAD_REQUEST,
            code=code,
            details=details,
        )


class BadLimitOffset(BadRequest):
    def __init__(self, details="Limit or offset out of range"):
        BadRequest.__init__(self, code="bad-limit-offset", details=details)


class BadFilter(BadRequest):
    def __init__(self, filter=None, details=None):
        if not details:
            if filter:
                details = f"Filter {filter} is not supported"
            else:
                details = "Filter is not supported"
        BadRequest.__init__(self, code="bad-filter", details=details)


class InvalidBarcode(BadRequest):
    def __init__(self, details="Invalid barcode"):
        BadRequest.__init__(self, code="invalid-barcode", details=details)


class UnsupportedMethod(BadRequest):
    def __init__(self, method=None, details=None):
        if not details:
            if method:
                details = f"Method {method} is not supported"
            else:
                details = "Method is not supported"
        BadRequest.__init__(self, code="unsupported-method", details=details)


class UnsupportedRequestAttr(BadRequest):
    def __init__(self, attr=None, details=None):
        if not details:
            if attr:
                details = f"Request attr {attr} is not supported"
            else:
                details = "Request attr is not supported"
        BadRequest.__init__(
            self, code="unsupported-request-attr", details=details
        )


class MethodNotAllowed(BaseApiException):
    def __init__(self, details="", code="method-not-allowed"):
        BaseApiException.__init__(
            self,
            status=falcon.HTTP_METHOD_NOT_ALLOWED,
            title=falcon.HTTP_METHOD_NOT_ALLOWED,
            code=code,
            details=details,
        )


class IdFieldMissingError(BadRequest):
    def __init__(
        self, details="Resource has no id field. This action is impossible."
    ):
        BadRequest.__init__(self, code="id-field-missing", details=details)



from dataclasses import dataclass, field
from typing import Dict, List, Optional

import sqlalchemy as sa
from marshmallow.fields import List as ListField
from sqlalchemy.orm import Session
from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy.sql.selectable import Select

from awokado.custom_fields import ToMany, ToOne
from awokado.exceptions import BadFilter, BadRequest, RelationNotFound
from awokado.filter_parser import filter_value_to_python, FilterItem
from awokado.utils import get_id_field, get_sort_way, has_resource_auth

if False:
    from awokado.resource import BaseResource


@dataclass
class ReadContext:
    session: Session
    resource: "BaseResource"
    uid: int

    # request vars
    include: Optional[List]
    query: Optional[List[FilterItem]]
    sort: Optional[list]
    resource_id: Optional[int]
    limit: Optional[int]
    offset: Optional[int]

    # runtime vars
    q: Select = field(default_factory=Select)
    obj_ids: List[int] = field(default_factory=list)
    parent_payload: List = field(default_factory=list)
    related_payload: Dict = field(default_factory=dict)
    total: int = 0

    # for Aggregation Resources
    time_scale = None

    @property
    def is_list(self):
        return not bool(self.resource_id)

    def read__filtering(self):
        if not self.query:
            return

        filters_to_apply = []

        for f in self.query:
            resource_field = self.resource.fields.get(f.field)

            # if not resource_field:
            #     raise BadFilter(
            #         details="Filed <{}> doesn't exist".format(f.field)
            #     )

            model_field = resource_field.metadata.get("model_field")

            if model_field is None:
                raise BadFilter(filter=f.field)

            value = f.wrapper(f.value)
            value = filter_value_to_python(value)

            list_deserialization = isinstance(value, list) and not isinstance(
                resource_field, ListField
            )

            if value is not None:
                if list_deserialization:
                    value = [resource_field.deserialize(item) for item in value]
                else:
                    value = resource_field.deserialize(value)

            field_expr = getattr(model_field, f.op)(value)
            filters_to_apply.append(field_expr)

        self.q = self.q.where(sa.and_(*filters_to_apply))

    def read__sorting(self):
        if self.sort:
            # TODO: add support for routes
            # example: (xxx.names.display_name)
            for sort_item in self.sort:
                sort_route, sort_way = get_sort_way(sort_item)

                if sort_route in self.resource.fields:
                    self.q = self.q.order_by(
                        sort_way(
                            self.resource.fields[sort_route].metadata.get(
                                "model_field"
                            )
                        )
                    )

    def read__pagination(self):
        if self.limit:
            self.q = self.q.limit(self.limit)
        if self.offset:
            self.q = self.q.offset(self.offset)

    def read__query(self):
        fields_to_select = {}
        to_group_by = []

        for field_name, field in self.resource.fields.items():
            model_field = field.metadata.get("model_field")
            if field.load_only:
                continue

            if model_field is None:
                raise Exception(
                    f"{self.resource.Meta.name}.{field_name} field must have "
                    f"'model_field' argument"
                )

            if isinstance(field, ToMany):
                fields_to_select[field_name] = sa.func.array_remove(
                    sa.func.array_agg(model_field), None
                )
            elif isinstance(field, ToOne):
                fields_to_select[field_name] = model_field
            else:
                fields_to_select[field_name] = model_field
                if (
                    isinstance(model_field, InstrumentedAttribute)
                    and model_field.class_ is not self.resource.Meta.model
                ):
                    to_group_by.append(model_field)

        q = sa.select([clm.label(lbl) for lbl, clm in fields_to_select.items()])

        joins = getattr(self.resource.Meta, "select_from", None)
        if joins is not None:
            q = q.select_from(joins)

        if not self.is_list:
            model_id = get_id_field(self.resource)
            q = q.where(model_id == self.resource_id)

        if has_resource_auth(self.resource):
            q = self.resource.Meta.auth.can_read(self, q)

        if joins is not None:
            model_id = get_id_field(self.resource)
            q = q.group_by(model_id, *to_group_by)

        self.q = q

    def __add_related_payload(self, related_res, related_data: list):
        if related_res.Meta.name in self.related_payload:
            related_res = related_res()
            related_res_id_field = get_id_field(related_res, name_only=True)
            existing_record_ids = [
                rec[related_res_id_field]
                for rec in self.related_payload[related_res.Meta.name]
            ]
            for rec in related_data:
                if rec[related_res_id_field] not in existing_record_ids:
                    self.related_payload[related_res.Meta.name].append(rec)
        else:
            self.related_payload[related_res.Meta.name] = related_data

    def read__includes(self):
        for i in self.include or []:
            if "." in i:
                raise BadRequest()

            elif not isinstance(self.resource.fields.get(i), (ToOne, ToMany)):
                raise RelationNotFound(f"Relation <{i}> not found")

            related_resource_name = self.resource.fields[i].metadata["resource"]
            related_res = self.resource.RESOURCES[related_resource_name]
            related_field = self.resource.fields[i].metadata.get("model_field")
            method_name = f"get_by_{self.resource.Meta.name.lower()}_ids"

            if not hasattr(related_res, method_name):
                raise BadRequest(
                    f"Relation {related_resource_name} doesn't ready yet. "
                    f"Ask developers to add {method_name} method "
                    f"to {related_resource_name} resource"
                )

            related_res_obj = related_res()
            related_data = getattr(related_res_obj, method_name)(
                self.session, self, related_field
            )
            self.__add_related_payload(related_res, related_data)

    def read__serializing(self) -> Dict:
        response = self.resource.Response(self.resource, self.is_list)
        response.set_parent_payload(self.parent_payload)
        response.set_related_payload(self.related_payload)
        response.set_total(self.total)
        serialized_response = response.serialize()
        return serialized_response

    def read__execute_query(self):
        if not self.resource.Meta.disable_total:
            self.q.append_column(sa.func.count().over().label("total"))

        result = self.session.execute(self.q).fetchall()

        serialized_data = self.resource.dump(result, many=True)

        id_field = get_id_field(self.resource, name_only=True)
        self.obj_ids.extend([_i[id_field] for _i in serialized_data])
        self.parent_payload = serialized_data

        if serialized_data and not self.resource.Meta.disable_total:
            self.total = result[0].total



import json

ERROR_STATUS = "status"
ERROR_CODE = "code"
ERROR_TITLE = "title"
ERROR_DETAIL = "detail"


class BaseApiException(Exception):
    def __init__(self, status, title, code=None, details="", headers=None):
        self.status = status
        self.title = title
        self.code = code
        self.details = details
        self.headers = headers

    def __repr__(self):
        return self.__str__()

    def has_representation(self):
        return True

    def __str__(self):
        details = ". {}".format(self.details) if self.details else self.details
        return "{status}{details}".format(status=self.status, details=details)

    def to_json(self):
        """
        :type self BaseApiException
        """
        obj = self.to_dict()
        return json.dumps(obj, ensure_ascii=False)

    def to_dict(self):
        """
        :type self BaseApiException
        """
        result = {ERROR_STATUS: self.status, ERROR_TITLE: self.title}
        if self.code:
            result[ERROR_CODE] = self.code
        if self.details:
            result[ERROR_DETAIL] = self.details
        return result

    @staticmethod
    def handle(ex, req, resp, params):

        raise ex



from typing import Dict

from sqlalchemy.sql import Selectable

from awokado.exceptions import (
    CreateResourceForbidden,
    ReadResourceForbidden,
    UpdateResourceForbidden,
    DeleteResourceForbidden,
)


class BaseAuth:
    """
    CREATE = {
    'ROLE NAME HERE': Boolean value
    }
            Example:
            'ADMIN': True,
            'GUEST': False

    READ,
    UPDATE,
    DELETE

    """

    CREATE: Dict[str, bool] = {
        # 'ROLE NAME HERE': Boolean value,
        #  Example:
        # 'ADMIN': True,
        # 'GUEST': False,
    }
    READ: Dict[str, bool] = {}
    UPDATE: Dict[str, bool] = {}
    DELETE: Dict[str, bool] = {}

    @classmethod
    def can_create(cls, session, payload, user_id: int, skip_exc=False):
        if skip_exc:
            return False
        raise CreateResourceForbidden()

    @classmethod
    def can_read(cls, ctx, query: Selectable, skip_exc=False):
        if skip_exc:
            return False

        raise ReadResourceForbidden()

    @classmethod
    def can_update(cls, session, user_id: int, obj_ids: list, skip_exc=False):
        if skip_exc:
            return False

        raise UpdateResourceForbidden()

    @classmethod
    def can_delete(cls, session, user_id: int, obj_ids: list, skip_exc=False):
        if skip_exc:
            return False

        raise DeleteResourceForbidden()

    @classmethod
    def _get_read_query(cls, ctx, query: Selectable):
        return query



import json
from typing import Any, Dict, List, Optional, Tuple, Union

import bulky
import falcon
import sqlalchemy as sa
from cached_property import cached_property
from clavis import Transaction
from marshmallow import utils, Schema, ValidationError
from marshmallow.schema import SchemaMeta

from awokado.auth import BaseAuth
from awokado.consts import (
    AUDIT_DEBUG,
    BULK_CREATE,
    BULK_UPDATE,
    CREATE,
    DELETE,
    OP_IN,
    UPDATE,
)
from awokado.custom_fields import ToMany, ToOne
from awokado.db import DATABASE_URL, persistent_engine
from awokado.exceptions import BadRequest, MethodNotAllowed
from awokado.filter_parser import FilterItem, OPERATORS_MAPPING
from awokado.request import ReadContext
from awokado.response import Response
from awokado.utils import (
    get_id_field,
    get_ids_from_payload,
    get_read_params,
    has_resource_auth,
)


class ResourceMeta(SchemaMeta):
    def __new__(mcs, cls_name=None, superclasses=None, attributes=None):
        new_resource = super(ResourceMeta, mcs).__new__(
            mcs, cls_name, superclasses, attributes
        )
        new_resource.RESOURCES[new_resource.Meta.name] = new_resource
        res_meta = new_resource.Meta

        if res_meta.name in ("base_resource", "_resource"):
            return new_resource

        if not getattr(res_meta, "methods", None):
            raise Exception(f"{cls_name} must have Meta.methods")
        if not getattr(res_meta, "name", None):
            raise Exception(f"{cls_name} must have Meta.name")

        new_resource_obj = new_resource()
        resource_id_name = get_id_field(
            new_resource_obj, name_only=True, skip_exc=True
        )
        if resource_id_name:
            resource_id_field = new_resource_obj.fields.get(resource_id_name)
            resource_id_field = resource_id_field.metadata.get("model_field")
            if not resource_id_field:
                raise Exception(
                    f"Resource's {cls_name} id field {resource_id_name}"
                    f" must have model_field."
                )

        if not hasattr(res_meta, "disable_total"):
            res_meta.disable_total = False

        return new_resource


class BaseResource(Schema, metaclass=ResourceMeta):
    RESOURCES: Dict[str, "BaseResource"] = {}
    Response = Response

    class Meta:
        """
        :param name:  used for two resources connection by relation
        :param model: represents sqlalchemy model or cte
        :param methods:  tuple of methods you want to allow
        :param auth: awokado `BaseAuth <reference.html#awokado.auth.BaseAuth>`_  class for embedding authentication logic
        :param skip_doc:  set true if you don't need to add the resource to documentation
        :param disable_total: set false, if you don't need to know returning objects amount in read-requests
        :param select_from: provide data source here if your resource use another's model fields (for example sa.outerjoin(FirstModel, SecondModel, FirstModel.id == SecondModel.first_model_id))
        """

        name = "base_resource"
        methods: Union[Tuple[str], Tuple] = tuple()
        model: Any = None  # type: ignore
        auth = BaseAuth
        skip_doc = True
        disable_total = False
        id_field = None

    ###########################################################################
    # Marshmallow validation methods
    ###########################################################################

    def validate_create_request(self, req: falcon.Request, is_bulk=False):
        methods = self.Meta.methods
        payload = json.load(req.bounded_stream)

        if isinstance(payload.get(self.Meta.name), list):
            request_method = BULK_CREATE
            is_bulk = True
        else:
            request_method = CREATE

        if request_method not in methods:
            raise MethodNotAllowed()

        data = payload.get(self.Meta.name)

        try:
            deserialized = self.load(data, many=is_bulk)
        except ValidationError as exc:
            if exc.messages == {"_schema": ["Invalid input type."]}:
                raise BadRequest(
                    f"Invalid schema, resource name is missing at the top level. "
                    f"Your POST request has to look like: "
                    f'{{"{self.Meta.name}": [{{"field_name": "field_value"}}] '
                    f'or {{"field_name": "field_value"}} }}'
                )
            raise BadRequest(exc.messages)

        try:
            deserialized = self.load(data, many=is_bulk)
        except ValidationError as exc:
            raise BadRequest(exc.messages)

        req.stream = {self.Meta.name: deserialized}

    def validate_update_request(self, req: falcon.Request):
        methods = self.Meta.methods
        if UPDATE not in methods and BULK_UPDATE not in methods:
            raise MethodNotAllowed()

        payload = json.load(req.bounded_stream)
        data = payload.get(self.Meta.name)
        try:
            deserialized = self.load(data, partial=True, many=True)
        except ValidationError as exc:
            raise BadRequest(exc.messages)

        req.stream = {self.Meta.name: deserialized}

    ###########################################################################
    # Falcon methods
    ###########################################################################

    def on_patch(
        self, req: falcon.Request, resp: falcon.Response, *args, **kwargs
    ):
        """
        Falcon method. PATCH-request entry point.

        Here is a database transaction opening.
        This is where authentication takes place (if auth class is pointed in `resource <reference.html#awokado.resource.BaseResource.Meta>`_)
        Then update method is run.

        :param req: falcon.Request
        :param resp: falcon.Response
        """
        with Transaction(DATABASE_URL, engine=persistent_engine) as t:
            session = t.session
            user_id, _ = self.auth(session, req, resp)

            self.validate_update_request(req)

            payload = req.stream

            data = payload[self.Meta.name]

            ids = get_ids_from_payload(self.Meta.model, data)

            if has_resource_auth(self):
                self.Meta.auth.can_update(session, user_id, ids)

            self.audit_log(
                f"Update: {self.Meta.name}", payload, user_id, AUDIT_DEBUG
            )

            result = self.update(session, payload, user_id)

        resp.body = json.dumps(result, default=str)

    def on_post(self, req: falcon.Request, resp: falcon.Response):
        """
        Falcon method. POST-request entry point.

        Here is a database transaction opening.
        This is where authentication takes place (if auth class is pointed in `resource <reference.html#awokado.resource.BaseResource.Meta>`_)
        Then create method is run.

        :param req: falcon.Request
        :param resp: falcon.Response
        """
        with Transaction(DATABASE_URL, engine=persistent_engine) as t:
            session = t.session
            user_id, token = self.auth(session, req, resp)

            self.validate_create_request(req)

            payload = req.stream

            if has_resource_auth(self):
                self.Meta.auth.can_create(
                    session, payload, user_id, skip_exc=False
                )

            self.audit_log(
                f"Create: {self.Meta.name}", payload, user_id, AUDIT_DEBUG
            )

            result = self.create(session, payload, user_id)

        resp.body = json.dumps(result, default=str)

    def auth(self, *args, **kwargs) -> Tuple[int, str]:
        """
        this method should return (user_id, token) tuple
        """
        return 0, ""

    def audit_log(self, *args, **kwargs):
        return

    def on_get(
        self,
        req: falcon.Request,
        resp: falcon.Response,
        resource_id: int = None,
    ):
        """
        Falcon method. GET-request entry point.

        Here is a database transaction opening.
        This is where authentication takes place (if auth class is pointed in `resource <reference.html#awokado.resource.BaseResource.Meta>`_)
        Then read_handler method is run. It's responsible for the whole read workflow.

        :param req: falcon.Request
        :param resp: falcon.Response
        """
        with Transaction(DATABASE_URL, engine=persistent_engine) as t:
            session = t.session
            user_id, token = self.auth(session, req, resp)
            params = get_read_params(req, self.__class__)
            params["resource_id"] = resource_id

            result = self.read_handler(session, user_id, **params)

        resp.body = json.dumps(result, default=str)

    def on_delete(
        self,
        req: falcon.Request,
        resp: falcon.Response,
        resource_id: int = None,
    ):
        """
        Falcon method. DELETE-request entry point.

        Here is a database transaction opening.
        This is where authentication takes place (if auth class is pointed in `resource <reference.html#awokado.resource.BaseResource.Meta>`_)
        Then delete method is run.

        :param req: falcon.Request
        :param resp: falcon.Response
        """

        with Transaction(DATABASE_URL, engine=persistent_engine) as t:
            session = t.session
            user_id, token = self.auth(session, req, resp)

            if DELETE not in self.Meta.methods:
                raise MethodNotAllowed()

            ids_to_delete = req.get_param_as_list("ids")

            data = [ids_to_delete, resource_id]
            if not any(data) or all(data):
                raise BadRequest(
                    details=(
                        "It should be a bulk delete (?ids=1,2,3) or delete"
                        " of a single resource (v1/resource/1)"
                    )
                )

            if not ids_to_delete:
                ids_to_delete = [resource_id]

            if has_resource_auth(self):
                self.Meta.auth.can_delete(session, user_id, ids_to_delete)

            result = self.delete(session, user_id, ids_to_delete)

        resp.body = json.dumps(result, default=str)

    ###########################################################################
    # Resource methods
    ###########################################################################

    def update(
        self, session, payload: dict, user_id: int, *args, **kwargs
    ) -> Dict:
        """

        First of all, data is prepared for updating:
        Marshmallow load method for data structure deserialization and then preparing data for SQLAlchemy update query.

        Updates data with bulk_update_mappings sqlalchemy method. Saves many-to-many relationships.

        Returns updated resources with the help of read_handler method.
        """
        data = payload[self.Meta.name]

        data_to_update = self._to_update(data)

        ids = get_ids_from_payload(self.Meta.model, data_to_update)

        session.bulk_update_mappings(self.Meta.model, data_to_update)
        self._save_m2m(session, data, update=True)

        op = OPERATORS_MAPPING[OP_IN]
        result = self.read_handler(
            session=session,
            user_id=user_id,
            filters=[FilterItem("id", op[0], op[1], ids)],
        )

        return result

    def create(self, session, payload: dict, user_id: int) -> dict:
        """

        Create method.

        You can override it to add your logic.

        First of all, data is prepared for creating:
        Marshmallow load method for data structure deserialization and then preparing data for SQLAlchemy create a query.

        Inserts data to the database (Uses bulky library if there is more than one entity to create). Saves many-to-many relationships.

        Returns created resources with the help of read_handler method.

        """
        # prepare data to insert
        data = payload[self.Meta.name]

        if isinstance(data, list):
            return self.bulk_create(session, user_id, data)

        data_to_insert = self._to_create(data)

        # insert to DB
        resource_id = session.execute(
            sa.insert(self.Meta.model)
            .values(data_to_insert)
            .returning(self.Meta.model.id)
        ).scalar()

        data["id"] = resource_id
        self._save_m2m(session, data)

        return self.read_handler(
            session=session, user_id=user_id, resource_id=resource_id
        )

    def bulk_create(self, session, user_id: int, data: list) -> dict:
        data_to_insert = []
        for item in data:
            data_to_insert.append(self._to_create(item))

        # insert to DB
        resource_ids = bulky.insert(
            session,
            self.Meta.model,
            data_to_insert,
            returning=[self.Meta.model.id],
        )
        ids = [r.id for r in resource_ids]

        op = OPERATORS_MAPPING[OP_IN]
        result = self.read_handler(
            session=session,
            user_id=user_id,
            filters=[FilterItem("id", op[0], op[1], ids)],
        )

        return result

    def delete(self, session, user_id: int, obj_ids: list):
        """

        Simply deletes objects with passed identifiers.

        """
        session.execute(
            sa.delete(self.Meta.model).where(self.Meta.model.id.in_(obj_ids))
        )
        return {}

    def _to_update(self, data: list) -> list:
        """
        Prepare resource data for SQLAlchemy update query
        """
        to_update_list = []
        for data_line in data:
            to_update = {}
            for fn, v in data_line.items():
                f = self.fields[fn]
                if isinstance(f, ToMany):
                    continue
                model_field = f.metadata.get("model_field")
                if not model_field:
                    continue
                to_update[model_field.key] = v

            to_update_list.append(to_update)

        return to_update_list

    def _to_create(self, data: dict) -> dict:
        """
        Prepare resource data for SQLAlchemy create query
        """
        to_create = {}
        for fn, v in data.items():
            f = self.fields[fn]
            if isinstance(f, ToMany):
                continue
            model_field = f.metadata["model_field"]
            to_create[model_field.key] = v

        return to_create

    def read_handler(
        self,
        session,
        user_id: int,
        include: list = None,
        filters: Optional[List[FilterItem]] = None,
        sort: list = None,
        resource_id: int = None,
        limit: int = None,
        offset: int = None,
    ) -> Dict:

        ctx = ReadContext(
            session,
            self,
            user_id,
            include,
            filters,
            sort,
            resource_id,
            limit,
            offset,
        )

        self.read__query(ctx)
        self.read__filtering(ctx)
        self.read__sorting(ctx)

        self.read__pagination(ctx)
        self.read__execute_query(ctx)

        if not ctx.obj_ids:
            if ctx.is_list:
                response = self.Response(self, is_list=ctx.is_list)
                return response.serialize()
            else:
                raise BadRequest("Object Not Found")

        self.read__includes(ctx)
        return self.read__serializing(ctx)

    def read__query(self, ctx: ReadContext):
        return ctx.read__query()

    def read__filtering(self, ctx: ReadContext):
        return ctx.read__filtering()

    def read__sorting(self, ctx: ReadContext):
        return ctx.read__sorting()

    def read__pagination(self, ctx: ReadContext):
        return ctx.read__pagination()

    def read__execute_query(self, ctx: ReadContext):
        return ctx.read__execute_query()

    def read__includes(self, ctx: ReadContext):
        return ctx.read__includes()

    def read__serializing(self, ctx: ReadContext) -> Dict:
        return ctx.read__serializing()

    def get_related_model(self, field: Union[ToOne, ToMany]):
        resource_name = field.metadata.get("resource")
        resource = self.RESOURCES[resource_name]
        return resource.Meta.model

    def _process_to_many_field(
        self, field_name: str, field: Union[ToOne, ToMany]
    ):
        related_model = self.get_related_model(field)
        resource_model = self.Meta.model
        model_field = field.metadata.get("model_field")

        if not isinstance(model_field, sa.Column):
            model_field = getattr(
                model_field.parent.persist_selectable.c, model_field.key
            )

        if related_model.__table__ == model_field.table:
            for fk in model_field.table.foreign_keys:
                if fk.column.table == resource_model.__table__:
                    setattr(field, "left_fk_field", fk.parent)
                    break
        else:
            setattr(field, "secondary", model_field.table)
            for fk in model_field.table.foreign_keys:
                if fk.column.table == related_model.__table__:
                    setattr(field, "right_fk_field", fk.parent)

                if fk.column.table == resource_model.__table__:
                    setattr(field, "left_fk_field", fk.parent)

        setattr(field, "related_model", related_model)
        return field_name, field

    @cached_property
    def _to_many_fields(self) -> list:
        return [
            self._process_to_many_field(field_name, field)
            for field_name, field in self.fields.items()
            if isinstance(field, ToMany)
        ]

    @staticmethod
    def check_exists(session, table: sa.Table, ids: list, field_name: str):
        result = session.execute(
            sa.select([table.c.id]).where(table.c.id.in_(ids))
        )

        missed = set(ids) - {item.id for item in result}
        if missed:
            err_msg = "objects with id {ids} does not exist".format(
                ids=",".join(map(str, missed))
            )
            raise BadRequest({field_name: err_msg})

    def _save_m2m(
        self, session, data: Union[list, dict], update: bool = False
    ) -> None:
        data = data if utils.is_collection(data) else [data]

        for field_name, field in self._to_many_fields:
            if hasattr(field, "secondary"):
                many_2_many = []
                for obj in data:
                    if field_name in obj:
                        many_2_many.extend(
                            [
                                {
                                    field.left_fk_field: obj.get("id"),
                                    field.right_fk_field: rel_id,
                                }
                                for rel_id in obj.get(field_name)
                            ]
                        )

                if update:
                    session.execute(
                        sa.delete(field.secondary).where(
                            field.left_fk_field.in_(
                                [obj.get("id") for obj in data]
                            )
                        )
                    )

                if many_2_many:
                    self.check_exists(
                        session,
                        field.related_model.__table__,
                        [obj.get(field.right_fk_field) for obj in many_2_many],
                        field_name,
                    )
                    session.execute(
                        sa.insert(field.secondary).values(many_2_many)
                    )



import sqlalchemy as sa
from sqlalchemy.pool import QueuePool

import clavis
from dynaconf import settings

from awokado.db_helper import Database


database = Database.from_config(settings)

DATABASE_URL = database.db_url
DATABASE_PASSWORD = database.DATABASE_PASSWORD
DATABASE_HOST = database.DATABASE_HOST
DATABASE_USER = database.DATABASE_USER
DATABASE_PORT = database.DATABASE_PORT
DATABASE_DB = database.DATABASE_DB

clavis.configure(DATABASE_URL)

persistent_engine = sa.create_engine(
    DATABASE_URL,
    encoding="utf-8",
    echo=settings.get("DB_ECHO", False),
    poolclass=QueuePool,
    pool_size=settings.get("DB_CONN_POOL_SIZE", 10),
    max_overflow=settings.get("DB_CONN_MAX_OVERFLOW", 5),
)



from .auth import (
    AuthError,
    BadEmail,
    BadPassword,
    BadPhone,
    BadPhoneVerificationCode,
    DataError,
    IdentificationFailed,
    PasswordMismatch,
    PasswordResetFailed,
    PhoneVerificationFailed,
    UserNotFound,
)
from .bad_request import (
    BadFilter,
    BadLimitOffset,
    BadRequest,
    InvalidBarcode,
    MethodNotAllowed,
    UnsupportedMethod,
    UnsupportedRequestAttr,
    IdFieldMissingError,
)
from .base import BaseApiException
from .forbidden import (
    CreateResourceForbidden,
    DeleteResourceForbidden,
    Forbidden,
    ReadResourceForbidden,
    UpdateResourceForbidden,
)
from .not_found import NotFound, RelationNotFound, ResourceNotFound



from typing import Dict, Optional, List


class Response:
    """
    Response class helps to collect your data
    and prepare it in a readable format for the Frontend (or another API Client)

    You can override it in your resource to change response format::

        class MyResponse(Response):
            PAYLOAD_KEYWORD = "data"

        class MyBaseResource(BaseResource):
            Response = MyResponse

    Default serialization for list requests (``/v1/book/``)::

        {
          "payload": {
            "book": [
              {
                "name": "My Book",
                "authors": [1, 2]
              }
            ]
          },
          "meta": {
            "total": 1
          }
        }

    Default serialization for single object (``/v1/book/123``)::

        {
          "book": [
            {
              "name": "My Book",
              "authors": [1, 2]
            }
          ]
        }

    """

    PAYLOAD_KEYWORD = "payload"
    META_KEYWORD = "meta"
    TOTAL_KEYWORD = "total"

    def __init__(self, resource, is_list: bool = False):
        """
        # :param resource: resource.BaseResource
        """
        self.is_list = is_list
        self.resource = resource

        self.payload: Dict = {}
        self.related_payload: Optional[Dict] = None
        self.include_total = False
        self.total = 0

        if resource:
            self.include_total = not resource.Meta.disable_total

    def serialize(self) -> Dict:
        if self.related_payload and self.payload:
            self.payload.update(self.related_payload)

        if self.is_list:
            return self._serialize_list()
        else:
            return self._serialize_single()

    def set_parent_payload(self, parent_payload: Optional[List] = None) -> None:
        if not parent_payload:
            parent_payload = []

        payload = {self.resource.Meta.name: parent_payload}
        self.payload = payload

    def set_related_payload(self, related_payload: Optional[Dict]) -> None:
        self.related_payload = related_payload

    def set_total(self, total_objects_count: int):
        self.total = total_objects_count

    def _serialize_single(self) -> Dict:
        if not self.payload:
            self.set_parent_payload()

        response: Dict = self.payload
        return response

    def _serialize_list(self) -> Dict:
        if not self.payload:
            self.set_parent_payload()

        response = {self.PAYLOAD_KEYWORD: self.payload}

        if self.include_total:
            response[self.META_KEYWORD] = {self.TOTAL_KEYWORD: self.total}
        else:
            response[self.META_KEYWORD] = None  # type: ignore

        return response



from marshmallow import fields, validate, utils


class ToMany(fields.List):
    def __init__(self, *args, **kwarg):
        related_resource_name = kwarg.get("resource")
        description = f"IDs of related resource ({related_resource_name})."

        users_description = kwarg.pop("description", None)
        if users_description:
            description = f"{description} {users_description}"

        super().__init__(*args, description=description, **kwarg)


class ToOne(fields.Integer):
    def __init__(self, *args, **kwarg):
        related_resource_name = kwarg.get("resource")
        description = f"ID of related resource ({related_resource_name})"

        users_description = kwarg.pop("description", None)
        if users_description:
            description = f"{description} {users_description}"

        super().__init__(*args, description=description, **kwarg)


class Choice(fields.Str):
    def __init__(self, *args, **kwarg):
        allowed_values = kwarg.get("allowed_values")
        super().__init__(
            *args, validate=validate.OneOf(allowed_values), **kwarg
        )



from contextlib import contextmanager
from dataclasses import dataclass
from logging import getLogger
from urllib.parse import urlparse

from sqlalchemy import create_engine, MetaData

log = getLogger("db_helper")


@dataclass
class Database:
    DATABASE_USER: str
    DATABASE_PASSWORD: str
    DATABASE_HOST: str
    DATABASE_DB: str
    DATABASE_PORT: str = "8123"

    @classmethod
    def from_url(cls, url: str):
        result = urlparse(url)

        data = {
            "DATABASE_USER": result.username,
            "DATABASE_PASSWORD": result.password,
            "DATABASE_DB": result.path[1:],
            "DATABASE_HOST": result.hostname,
        }

        if result.port:
            data["DATABASE_PORT"] = str(result.port)

        if result.scheme != "postgresql" or not all(data.values()):
            raise Exception(f"Invalid url: {url}")

        return cls(**data)

    @classmethod
    def from_config(cls, config):
        url = config.get("DATABASE_URL", None)

        db_params = {
            "DATABASE_PASSWORD": config.get("DATABASE_PASSWORD", None),
            "DATABASE_HOST": config.get("DATABASE_HOST", None),
            "DATABASE_USER": config.get("DATABASE_USER", None),
            "DATABASE_DB": config.get("DATABASE_DB", None),
        }

        port = config.get("DATABASE_PORT", None)
        if port:
            db_params["DATABASE_PORT"] = str(port)

        if any(db_params.values()):
            if url:
                raise Exception(
                    "Cannot specify DATABASE_URL and any of other database params"
                    " together! (Other database params are: DATABASE_PASSWORD,"
                    " DATABASE_HOST, DATABASE_USER, DATABASE_PORT, DATABASE_DB)"
                )
            return cls(**db_params)
        elif url:
            return cls.from_url(url)
        else:
            raise Exception(
                "Please, specify DATABASE_URL or database params "
                "(DATABASE_PASSWORD, DATABASE_HOST, DATABASE_USER, DATABASE_PORT, DATABASE_DB)"
            )

    @property
    def url(self) -> str:
        if self.DATABASE_PORT:
            host = f"{self.DATABASE_HOST}:{self.DATABASE_PORT}"
        else:
            host = f"{self.DATABASE_HOST}"

        return (
            f"postgresql://{self.DATABASE_USER}:{self.DATABASE_PASSWORD}@{host}"
        )

    @property
    def db_url(self) -> str:
        return f"{self.url}/{self.DATABASE_DB}"

    @contextmanager
    def conn(self):
        e = create_engine(self.url)
        conn = e.connect()
        try:
            yield conn
        finally:
            conn.close()

    def drop(self):
        log.info(f"Drop database: {self.DATABASE_DB}")
        with self.conn() as conn:
            conn.execute("commit")
            conn.execute(f"drop database if exists {self.DATABASE_DB}")
            conn.execute("commit")

    def create(self):
        log.info(f"Create database: {self.DATABASE_DB}")
        with self.conn() as conn:
            conn.execute("commit")
            conn.execute(f"create database {self.DATABASE_DB}")
            conn.execute("commit")

    def recreate(self):
        self.drop()
        self.create()

    def create_models(self, md: MetaData):
        log.info(f"Create models: {self.DATABASE_DB}")
        e = create_engine(self.db_url)
        md.create_all(e)

    def __str__(self):
        return self.db_url



from collections import OrderedDict
from decimal import Decimal

import falcon
import pyaml
from apispec import APISpec
from apispec.ext.marshmallow import MarshmallowPlugin
from apispec.ext.marshmallow.openapi import DEFAULT_FIELD_MAPPING

from awokado import custom_fields
from awokado.consts import BULK_UPDATE, BULK_CREATE
from awokado.documentation.readme import get_readme
from awokado.documentation.routes import collect_routes
from awokado.documentation.utils import parse_doc_string
from awokado.resource import BaseResource
from awokado.utils import log

DEFAULT_FIELD_MAPPING.update(
    {
        custom_fields.ToOne: ("integer", None),
        custom_fields.ToMany: ("array", None),
        custom_fields.Choice: ("string", "string"),
    }
)

INNER_METHOD_MAP = {
    "patch": "update",
    "post": "create",
    "get": "on_get",  # not taking here
    "delete": "delete",
}


class APIDocs:
    def __init__(
        self,
        routes: list,
        models: dict,
        host,
        public_resources: frozenset = frozenset(),
    ):
        self.doc = models
        self.routes = routes
        self.host = host
        self.public_resources = public_resources

    def collect_path_info(self, r):
        result = {}
        resource_ext = r.route.split("/")[2]
        resource_name = resource_ext.replace("-", "_")
        route_method = r.method_name.lower()
        is_patch = route_method == "patch"
        is_post = route_method == "post"
        is_bulk = False
        route_w_id = "resource_id" in r.route

        if is_patch:
            is_bulk = BULK_UPDATE in r.resource.Meta.methods and not route_w_id

        if not route_w_id and is_patch and not is_bulk:
            log.info(
                f"skip bulk patch with resource id "
                f"{r.route}:{resource_name}:{route_method}"
            )
            return None, None

        if route_w_id and is_post:
            log.info(
                f"skip post with resource id "
                f"{r.route}:{resource_name}:{route_method}"
            )
            return None, None

        path_info = self.get_common_info(r, resource_name)
        path_info[r.route][r.method_name]["parameters"] = self.get_parameters(
            route_method, route_w_id, resource_name
        )

        if is_patch or is_post:
            rb = OrderedDict(
                [
                    ("required", True),
                    (
                        "content",
                        {
                            "application/json": {
                                "schema": {
                                    "$ref": (
                                        "#/components/schemas/" + resource_name
                                    )
                                }
                            }
                        },
                    ),
                ]
            )
            path_info[r.route][r.method_name]["requestBody"] = rb

        if resource_name not in self.public_resources:
            security = path_info[r.route][r.method_name]["security"] = []
            security.append({"bearerAuth": []})
        if isinstance(result.get(r.route), dict):
            result[r.route].update(path_info[r.route])
        else:
            result.update(path_info)
        return r.route, result[r.route]

    def get_common_info(self, r, name):
        inner_method = INNER_METHOD_MAP[r.method_name.lower()]
        method_doc_str = getattr(r.resource, inner_method).__doc__
        if not method_doc_str:
            method_doc_str = r.method.__doc__

        summary, description = parse_doc_string(method_doc_str)

        if r.method_name == "post" and BULK_CREATE in r.resource.Meta.methods:
            summary += ". Supports bulk create"

        if description:
            description = description.replace("---", "")

        path_info = {
            r.route: {
                r.method_name: {
                    "summary": summary,
                    "description": description,
                    "responses": {
                        200: {
                            "description": "OK",
                            "content": {
                                "application/json": {
                                    "schema": {
                                        "$ref": ("#/components/schemas/" + name)
                                    }
                                }
                            },
                        }
                    },
                }
            }
        }
        return path_info

    def get_parameters(self, route_method, route_with_id, resource_name):
        parameters = []
        if route_with_id:
            p = OrderedDict(
                [
                    ("name", "resource_id"),
                    ("in", "path"),
                    ("description", "ID of resource"),
                    ("required", True),
                    (
                        "schema",
                        OrderedDict([("type", "integer"), ("format", "int64")]),
                    ),
                ]
            )
            parameters.append(p)
        else:
            if route_method == "get":
                params_description = (
                    ("limit", "The numbers of items to return", "integer"),
                    (
                        "offset",
                        "The number of items to skip before "
                        "starting to collect the result set",
                        "integer",
                    ),
                    (
                        "sort",
                        "Sorting fields. Use '-' before field name "
                        "to use DESC sorting example: full_name,-birth_date",
                        "string",
                    ),
                )
                for name, description, q_type in params_description:
                    p = OrderedDict(
                        [
                            ("name", f"{name}"),
                            ("in", "query"),
                            ("description", f"{description}"),
                            ("schema", {"type": f"{q_type}"}),
                        ]
                    )
                    parameters.append(p)

        return parameters

    def set_servers(self):
        self.doc["servers"] = [{"url": self.host}]

    def set_components(self):
        if "components" in self.doc:
            self.doc["components"]["securitySchemes"] = {
                "bearerAuth": {
                    "type": "http",
                    "scheme": "bearer",
                    "bearerFormat": "JWT",
                }
            }
        else:
            self.doc["components"] = {
                "securitySchemes": {
                    "bearerAuth": {
                        "type": "http",
                        "scheme": "bearer",
                        "bearerFormat": "JWT",
                    }
                }
            }

    def run(self):
        for r in self.routes:
            route_key, route_result = self.collect_path_info(r)

            if not route_result:
                continue

            if route_key not in self.doc["paths"]:
                self.doc["paths"][route_key] = route_result
            else:
                self.doc["paths"][route_key].update(route_result)

        self.set_components()
        self.set_servers()
        return self.doc


def decimal_presenter(dumper, data):
    return dumper.represent_scalar("tag:yaml.org,2002:str", str(data))


def generate_documentation(
    api: falcon.API,
    api_host: str,
    project_name: str,
    output_dir: str,
    api_version: str = "1.0.0",
    template_absolute_path: str = None,
):
    public_resources = set()
    spec = APISpec(
        title=project_name,
        version=api_version,
        openapi_version="3.0.2",
        info=dict(description=get_readme(template_absolute_path)),
        plugins=[MarshmallowPlugin()],
    )

    for resource_name, resource in BaseResource.RESOURCES.items():
        if hasattr(resource.Meta, "skip_doc") and resource.Meta.skip_doc:
            continue

        if getattr(resource.Meta, "auth", None) is None:
            public_resources.add(resource_name)

        spec.components.schema(resource_name, schema=resource)

    routes = collect_routes(api)
    models_definition = spec.to_dict()
    d = APIDocs(
        routes, models_definition, api_host, frozenset(public_resources)
    )
    data = d.run()

    pyaml.add_representer(Decimal, decimal_presenter)

    output_dir = output_dir.rstrip("/")

    with open(f"{output_dir}/swagger.yaml", "w") as f:
        result = pyaml.dump(data, safe=True)
        f.write(result)



from importlib.machinery import SourceFileLoader
from os import path

from setuptools import setup


VERSION = SourceFileLoader(
    "version", path.join(".", "awokado", "version.py")
).load_module()

VERSION = VERSION.__version__

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="awokado",
    version=VERSION,
    description="Fast and flexible API framework based on Falcon and SQLAlchemy",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/5783354/awokado",
    author="Dmitry Karnei",
    author_email="5783354@gmail.com",
    classifiers=(
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
    ),
    keywords=" ".join(
        sorted(
            {"api", "rest", "wsgi", "falcon", "sqlalchemy", "sqlalchemy-core"}
        )
    ),
    packages=["awokado", "awokado.documentation", "awokado.exceptions"],
    install_requires=(
        "bcrypt",
        "bulky",
        "boto3",
        "cached-property",
        "dynaconf",
        "falcon==2.0.0",
        "marshmallow>=3.0.0rc5",
        "pyaml",
        "clavis",
        "apispec==2.0.1",
        "jinja2",
        "SQLAlchemy>=1.3.0",
        "m2r",
    ),
    python_requires=">=3.7",
)



import datetime
import hashlib
import json
import logging
import random
import string
import sys
import traceback
import uuid
from json import JSONDecodeError
from typing import NamedTuple, List, Dict, Any

import falcon
from dynaconf import settings
from sqlalchemy import desc, asc

from awokado.consts import DEFAULT_ACCESS_CONTROL_HEADERS
from awokado.exceptions import BaseApiException, IdFieldMissingError, BadRequest
from awokado.filter_parser import parse_filters

log = logging.getLogger("awokado")


class AuthBundle(NamedTuple):
    user_id: int
    auth_token: str


def rand_string(size=8, chars=string.ascii_uppercase + string.digits):
    return "".join(random.choice(chars) for _ in range(size))


def get_sort_way(sort_route):
    if sort_route.startswith("-"):
        sort_route = sort_route[1:]

        def sort_way(sort_field):
            return desc(sort_field).nullslast()

    else:

        def sort_way(sort_field):
            return asc(sort_field).nullsfirst()

    return sort_route, sort_way


def empty_response(resource, is_list=False):
    if isinstance(resource, str):
        resource_name = resource
    else:
        resource_name = resource.Meta.name

    if is_list:
        response = {"payload": {resource_name: []}, "meta": None}
        if not resource.Meta.disable_total:
            response["meta"] = {"total": 0}
        return response

    else:
        return {resource_name: []}


def get_read_params(req, resource) -> dict:
    """
    :param req: falcon.Request
    :param resource: api.resources.base.BaseResource
    """

    params = dict(
        include=req.get_param_as_list("include"),
        sort=req.get_param_as_list("sort"),
        limit=req.get_param_as_int("limit"),
        offset=req.get_param_as_int("offset"),
    )

    params["filters"] = parse_filters(req._params, resource)
    return params


def has_resource_auth(resource):
    if not hasattr(resource.Meta, "auth"):
        return False
    if resource.Meta.auth is None:
        return False

    return True


def json_error_serializer(
    req: falcon.Request, resp: falcon.Response, exception: BaseApiException
):
    # Serialize exception
    resp.body = exception.to_json()

    # Set content type
    resp.content_type = "application/json"
    resp.append_header("Vary", "Accept")
    resp.status = exception.status

    # Setup CORS
    origin = req.headers.get("HTTP_ORIGIN")
    origin2 = req.headers.get("ORIGIN")
    origin = origin2 or origin
    headers = {}

    if settings.get("AWOKADO_DEBUG"):
        headers["Access-Control-Allow-Origin"] = origin
    else:
        if origin and origin in settings.ORIGIN_HOSTS:
            headers["Access-Control-Allow-Origin"] = origin

    headers_to_set = settings.get(
        "AWOKADO_ACCESS_CONTROL_HEADERS", DEFAULT_ACCESS_CONTROL_HEADERS
    )
    for k, v in headers_to_set:
        headers[k] = v

    resp.set_headers(headers)


def api_exception_handler(error, req, resp, params):
    if isinstance(error, BaseApiException):
        resp.status = error.status

        if error.headers is not None:
            resp.set_headers(error.headers)

        if error.has_representation:
            json_error_serializer(req, resp, error)

        if settings.get("AWOKADO_LOG_USERS_EXCEPTIONS", False):
            exc_info = sys.exc_info()
            log.error("User error: ", exc_info=exc_info)

    elif isinstance(error, falcon.HTTPNotFound):
        resp.status = "404 Not Found"
        resp.content_type = "application/json"
        resp.body = falcon.json.dumps({"error": f"{req.path} not found"})
        resp.append_header("Vary", "Accept")

    else:
        resp.status = "500 Internal Server Error"
        exc_info = sys.exc_info()
        log.error("api_exception_handler", exc_info=exc_info)

        if settings.get("AWOKADO_DEBUG"):

            if hasattr(error, "to_dict"):
                resp.body = falcon.json.dumps({"error": error.to_dict()})

            elif hasattr(error, "to_json"):
                json_data = error.to_json()

                try:
                    json_data = json.loads(json_data)
                except (TypeError, JSONDecodeError):
                    json_data = json_data

                resp.body = falcon.json.dumps({"error": json_data})

            else:
                exc_data = "".join(traceback.format_exception(*sys.exc_info()))
                resp.body = falcon.json.dumps({"error": exc_data})

        else:
            resp.body = falcon.json.dumps({"error": resp.status})

        # Set content type
        resp.content_type = "application/json"
        resp.append_header("Vary", "Accept")


def get_uuid_hash(string):
    return hashlib.md5(
        (
            str(uuid.uuid3(uuid.uuid4(), string))
            + str(datetime.datetime.utcnow())
        ).encode("utf-8")
    ).hexdigest()


def get_id_field(resource, name_only=False, skip_exc=False):
    resource_id_field = getattr(resource.Meta, "id_field", "id")
    if resource_id_field not in resource.fields:
        if skip_exc:
            return False

        raise IdFieldMissingError()

    if name_only:
        return resource_id_field

    resource_id_field = resource.fields.get(resource_id_field)
    resource_id_field = resource_id_field.metadata.get("model_field")
    return resource_id_field


def get_ids_from_payload(model: Any, payload: List[Dict]) -> List:
    if model and hasattr(model, "id"):
        ids = [d.get(model.id.key) for d in payload]
    else:
        raise BadRequest("Model's ID field doesn't specified")

    return ids
